# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Routines for setting up Fortran, common to all dialects."""

from __future__ import annotations

import re
import os.path

import SCons.Scanner.Fortran
import SCons.Tool
import SCons.Util
from SCons.Action import Action, CommandAction
from SCons.Defaults import StaticObjectEmitter, SharedObjectEmitter


def isfortran(env, source) -> bool:
    """Returns True if *source* has any fortran files in it.

    Only checks based on filename suffixes, does not examine code.
    """
    try:
        fsuffixes = env['FORTRANSUFFIXES']
    except KeyError:
        # If no FORTRANSUFFIXES, no fortran tool, so there is no need to look
        # for fortran sources.
        return False

    if not source:
        # Source might be None for unusual cases like SConf.
        return False
    for s in source:
        if s.sources:
            ext = os.path.splitext(str(s.sources[0]))[1]
            if ext in fsuffixes:
                return True
    return False


def _fortranEmitter(target, source, env) -> Tuple:
    """Common code for Fortran emitter.

    Called by both the static and shared object emitters,
    mainly to account for generated module files.
    """
    node = source[0].rfile()
    if not node.exists() and not node.is_derived():
        print("Could not locate " + str(node.name))
        return [], []

    # This has to match the def_regex in the Fortran scanner
    mod_regex = r"""(?i)^\s*MODULE\s+(?!PROCEDURE|SUBROUTINE|FUNCTION|PURE|ELEMENTAL)(\w+)"""
    cre = re.compile(mod_regex, re.M)
    # Retrieve all USE'd module names
    modules = cre.findall(node.get_text_contents())
    # Remove unique items from the list
    modules = SCons.Util.unique(modules)
    # Convert module name to a .mod filename
    suffix = env.subst('$FORTRANMODSUFFIX', target=target, source=source)
    moddir = env.subst('$FORTRANMODDIR', target=target, source=source)
    modules = [mod.lower() + suffix for mod in modules]
    for module in modules:
        target.append(env.fs.File(module, moddir))
    return target, source


def FortranEmitter(target, source, env) -> Tuple:
    """Create emitter for static objects."""
    target, source = _fortranEmitter(target, source, env)
    return StaticObjectEmitter(target, source, env)


def ShFortranEmitter(target, source, env) -> Tuple:
    """Create emitter for shared objects."""
    target, source = _fortranEmitter(target, source, env)
    return SharedObjectEmitter(target, source, env)


def ComputeFortranSuffixes(suffixes: list[str], ppsuffixes: list[str]) -> None:
    """Update the suffix lists to reflect the platform requirements.

    If upper-cased suffixes can be distinguished from lower, those are
    added to *ppsuffixes*. If not, they are added to *suffixes*.

    Args:
        suffixes: regular Fortran source files
        ppsuffixes: Fortran source files that should be
          be run through the pre-processor
    """
    assert len(suffixes) > 0
    s = suffixes[0]
    sup = s.upper()
    upper_suffixes = [suf.upper() for suf in suffixes]
    if SCons.Util.case_sensitive_suffixes(s, sup):
        ppsuffixes.extend(upper_suffixes)
    else:
        suffixes.extend(upper_suffixes)


def CreateDialectActions(
    dialect: str,
) -> tuple[CommandAction, CommandAction, CommandAction, CommandAction]:
    """Create dialect specific actions."""
    CompAction = Action(f'${dialect}COM ', cmdstr=f'${dialect}COMSTR')
    CompPPAction = Action(f'${dialect}PPCOM ', cmdstr=f'${dialect}PPCOMSTR')
    ShCompAction = Action(f'$SH{dialect}COM ', cmdstr=f'$SH{dialect}COMSTR')
    ShCompPPAction = Action(f'$SH{dialect}PPCOM ', cmdstr=f'$SH{dialect}PPCOMSTR')
    return CompAction, CompPPAction, ShCompAction, ShCompPPAction


def DialectAddToEnv(
    env,
    dialect: str,
    suffixes: list[str],
    ppsuffixes: list[str],
    support_mods: bool = False,
) -> None:
    """Add dialect specific construction variables.

    Args:
        dialect: dialect name
        suffixes: suffixes associated with this dialect
        ppsuffixes: suffixes using cpp associated with this dialect
        support_mods: whether this dialect supports modules
    """
    ComputeFortranSuffixes(suffixes, ppsuffixes)

    fscan = SCons.Scanner.Fortran.FortranScan(f"{dialect}PATH")
    for suffix in suffixes + ppsuffixes:
        SCons.Tool.SourceFileScanner.add_scanner(suffix, fscan)

    env.AppendUnique(FORTRANSUFFIXES=suffixes + ppsuffixes)

    compaction, compppaction, shcompaction, shcompppaction = \
            CreateDialectActions(dialect)
    static_obj, shared_obj = SCons.Tool.createObjBuilders(env)

    for suffix in suffixes:
        static_obj.add_action(suffix, compaction)
        shared_obj.add_action(suffix, shcompaction)
        static_obj.add_emitter(suffix, FortranEmitter)
        shared_obj.add_emitter(suffix, ShFortranEmitter)

    for suffix in ppsuffixes:
        static_obj.add_action(suffix, compppaction)
        shared_obj.add_action(suffix, shcompppaction)
        static_obj.add_emitter(suffix, FortranEmitter)
        shared_obj.add_emitter(suffix, ShFortranEmitter)

    if f'{dialect}FLAGS' not in env:
        env[f'{dialect}FLAGS'] = SCons.Util.CLVar('')
    if f'SH{dialect}FLAGS' not in env:
        env[f'SH{dialect}FLAGS'] = SCons.Util.CLVar(f'${dialect}FLAGS')

    # If a tool does not define fortran prefix/suffix for include path, use C ones
    if f'INC{dialect}PREFIX' not in env:
        env[f'INC{dialect}PREFIX'] = '$INCPREFIX'
    if f'INC{dialect}SUFFIX' not in env:
        env[f'INC{dialect}SUFFIX'] = '$INCSUFFIX'

    env[f'_{dialect}INCFLAGS'] = f'${{_concat(INC{dialect}PREFIX, {dialect}PATH, INC{dialect}SUFFIX, __env__, RDirs, TARGET, SOURCE, affect_signature=False)}}'

    if support_mods:
        env[f'{dialect}COM'] = f'${dialect} -o $TARGET -c $FORTRANCOMMONFLAGS ${dialect}FLAGS $_{dialect}INCFLAGS $_FORTRANMODFLAG $SOURCES'
        env[f'{dialect}PPCOM'] = f'${dialect} -o $TARGET -c $FORTRANCOMMONFLAGS ${dialect}FLAGS $CPPFLAGS $_CPPDEFFLAGS $_{dialect}INCFLAGS $_FORTRANMODFLAG $SOURCES'
        env[f'SH{dialect}COM'] = f'$SH{dialect} -o $TARGET -c $FORTRANCOMMONFLAGS $SH{dialect}FLAGS $_{dialect}INCFLAGS $_FORTRANMODFLAG $SOURCES'
        env[f'SH{dialect}PPCOM'] = f'$SH{dialect} -o $TARGET -c $FORTRANCOMMONFLAGS $SH{dialect}FLAGS $CPPFLAGS $_CPPDEFFLAGS $_{dialect}INCFLAGS $_FORTRANMODFLAG $SOURCES'
    else:
        env[f'{dialect}COM'] = f'${dialect} -o $TARGET -c $FORTRANCOMMONFLAGS ${dialect}FLAGS $_{dialect}INCFLAGS $SOURCES'
        env[f'{dialect}PPCOM'] = f'${dialect} -o $TARGET -c $FORTRANCOMMONFLAGS ${dialect}FLAGS $CPPFLAGS $_CPPDEFFLAGS $_{dialect}INCFLAGS $SOURCES'
        env[f'SH{dialect}COM'] = f'$SH{dialect} -o $TARGET -c $FORTRANCOMMONFLAGS $SH{dialect}FLAGS $_{dialect}INCFLAGS $SOURCES'
        env[f'SH{dialect}PPCOM'] = f'$SH{dialect} -o $TARGET -c $FORTRANCOMMONFLAGS $SH{dialect}FLAGS $CPPFLAGS $_CPPDEFFLAGS $_{dialect}INCFLAGS $SOURCES'


def add_fortran_to_env(env) -> None:
    """Add Builders and construction variables for Fortran/generic."""
    FortranSuffixes = env.get('FORTRANFILESUFFIXES', ['.f', '.for', '.ftn'])
    FortranPPSuffixes = env.get('FORTRANPPFILESUFFIXES', ['.fpp', '.FPP'])
    DialectAddToEnv(env, "FORTRAN", FortranSuffixes, FortranPPSuffixes, support_mods=True)

    # Module support
    env['FORTRANMODPREFIX'] = ''     # like $LIBPREFIX
    env['FORTRANMODSUFFIX'] = '.mod' # like $LIBSUFFIX
    env['FORTRANMODDIR'] = ''          # where the compiler should place .mod files
    env['FORTRANMODDIRPREFIX'] = ''    # some prefix to $FORTRANMODDIR - similar to $INCPREFIX
    env['FORTRANMODDIRSUFFIX'] = ''    # some suffix to $FORTRANMODDIR - similar to $INCSUFFIX
    env['_FORTRANMODFLAG'] = '$( ${_concat(FORTRANMODDIRPREFIX, FORTRANMODDIR, FORTRANMODDIRSUFFIX, __env__, RDirs, TARGET, SOURCE)} $)'

def add_f77_to_env(env) -> None:
    """Add Builders and construction variables for f77 dialect."""
    F77Suffixes = env.get('F77FILESUFFIXES', ['.f77'])
    F77PPSuffixes = env.get('F77PPFILESUFFIXES', [])
    DialectAddToEnv(env, "F77", F77Suffixes, F77PPSuffixes)

def add_f90_to_env(env) -> None:
    """Add Builders and construction variables for f90 dialect."""
    F90Suffixes = env.get('F90FILESUFFIXES', ['.f90'])
    F90PPSuffixes = env.get('F90PPFILESUFFIXES', [])
    DialectAddToEnv(env, "F90", F90Suffixes, F90PPSuffixes, support_mods=True)

def add_f95_to_env(env) -> None:
    """Add Builders and construction variables for f95 dialect."""
    F95Suffixes = env.get('F95FILESUFFIXES', ['.f95'])
    F95PPSuffixes = env.get('F95PPFILESUFFIXES', [])
    DialectAddToEnv(env, "F95", F95Suffixes, F95PPSuffixes, support_mods=True)

def add_f03_to_env(env) -> None:
    """Add Builders and construction variables for f03 dialect."""
    F03Suffixes = env.get('F03FILESUFFIXES', ['.f03'])
    F03PPSuffixes = env.get('F03PPFILESUFFIXES', [])
    DialectAddToEnv(env, "F03", F03Suffixes, F03PPSuffixes, support_mods=True)

def add_f08_to_env(env) -> None:
    """Add Builders and construction variables for f08 dialect."""
    F08Suffixes = env.get('F08FILESUFFIXES', ['.f08'])
    F08PPSuffixes = env.get('F08PPFILESUFFIXES', [])
    DialectAddToEnv(env, "F08", F08Suffixes, F08PPSuffixes, support_mods=True)

def add_all_to_env(env) -> None:
    """Add builders and construction variables for all supported dialects."""
    add_fortran_to_env(env)
    add_f77_to_env(env)
    add_f90_to_env(env)
    add_f95_to_env(env)
    add_f03_to_env(env)
    add_f08_to_env(env)

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import re

_is_valid_var = re.compile(r'[_a-zA-Z]\w*$')

_rm = re.compile(r'\$[()]')
_remove = re.compile(r'\$\([^$]*(\$[^)][^$]*)*\$\)')

# Regular expressions for splitting strings and handling substitutions,
# for use by the scons_subst() and scons_subst_list() functions:
#
# The first expression compiled matches all of the $-introduced tokens
# that we need to process in some way, and is used for substitutions.
# The expressions it matches are:
#
#       "$$"
#       "$("
#       "$)"
#       "$variable"             [must begin with alphabetic or underscore]
#       "${any stuff}"
#
# The second expression compiled is used for splitting strings into tokens
# to be processed, and it matches all of the tokens listed above, plus
# the following that affect how arguments do or don't get joined together:
#
#       "   "                   [white space]
#       "non-white-space"       [without any dollar signs]
#       "$"                     [single dollar sign]
#
_dollar_exps_str = r'\$[\$\(\)]|\$[_a-zA-Z][\.\w]*|\${[^}]*}'
_dollar_exps = re.compile(r'(%s)' % _dollar_exps_str)
_separate_args = re.compile(r'(%s|\s+|[^\s$]+|\$)' % _dollar_exps_str)

# This regular expression is used to replace strings of multiple white
# space characters in the string result from the scons_subst() function.
_space_sep = re.compile(r'[\t ]+(?![^{]*})')

class ValueTypes:
    """
    Enum to store what type of value the variable holds.
    """
    UNKNOWN = 0
    STRING = 1
    CALLABLE = 2
    VARIABLE = 3


class EnvironmentValue:
    """
    Hold a single value. We're going to cache parsed version of the file
    We're going to keep track of variables which feed into this values evaluation
    """
    def __init__(self, value) -> None:
        self.value = value
        self.var_type = ValueTypes.UNKNOWN

        if callable(self.value):
            self.var_type = ValueTypes.CALLABLE
        else:
            self.parse_value()


    def parse_value(self) -> None:
        """
        Scan the string and break into component values
        """

        try:
            if '$' not in self.value:
                self._parsed = self.value
                self.var_type = ValueTypes.STRING
            else:
                # Now we need to parse the specified string
                result = _dollar_exps.sub(sub_match, args)
                print(result)
        except TypeError:
            # likely callable? either way we don't parse
            self._parsed = self.value

    def parse_trial(self) -> None:
        """
        Try alternate parsing methods.
        :return:
        """
        parts = []
        for c in self.value:
            pass


class EnvironmentValues:
    """
    A class to hold all the environment variables
    """
    def __init__(self, **kw) -> None:
        self._dict = {}
        for k in kw:
            self._dict[k] = EnvironmentValue(kw[k])



"""SCons.Tool.mslib

Tool-specific initialization for lib (MicroSoft library archiver).

There normally shouldn't be any need to import this module directly.
It will usually be imported through the generic SCons.Tool.Tool()
selection method.

"""

#
# __COPYRIGHT__
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "__FILE__ __REVISION__ __DATE__ __DEVELOPER__"

import os

import SCons.Defaults
import SCons.Tool
import SCons.Tool.msvs
import SCons.Tool.msvc
import SCons.Util

from SCons.Tool.MSCommon import (
    msvc_setup_env_tool,
    msvc_setup_env_once,
)

tool_name = 'mslib'

def generate(env) -> None:
    """Add Builders and construction variables for lib to an Environment."""
    SCons.Tool.createStaticLibBuilder(env)

    # Set-up ms tools paths
    msvc_setup_env_once(env, tool=tool_name)

    env['AR']          = 'lib'
    env['ARFLAGS']     = SCons.Util.CLVar('/nologo')
    env['ARCOM']       = "${TEMPFILE('$AR $ARFLAGS /OUT:$TARGET $SOURCES','$ARCOMSTR')}"
    env['LIBPREFIX']   = ''
    env['LIBSUFFIX']   = '.lib'

    # Issue #3350
    # Change tempfile argument joining character from a space to a newline
    # mslink will fail if any single line is too long, but is fine with many lines
    # in a tempfile
    env['TEMPFILEARGJOIN'] = os.linesep


def exists(env):
    return msvc_setup_env_tool(env, tool=tool_name)

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from SCons.Tool import createLoadableModuleBuilder
from .SharedLibrary import shlib_symlink_emitter
from . import lib_emitter


def ldmod_symlink_emitter(target, source, env, **kw):
    return shlib_symlink_emitter(target, source, env, variable_prefix='LDMODULE')


def _get_ldmodule_stem(target, source, env, for_signature):
    """
    Get the basename for a library (so for libxyz.so, return xyz)
    :param target:
    :param source:
    :param env:
    :param for_signature:
    :return:
    """
    target_name = str(target)
    ldmodule_prefix = env.subst('$LDMODULEPREFIX')
    ldmodule_suffix = env.subst("$_LDMODULESUFFIX")

    if target_name.startswith(ldmodule_prefix):
        target_name = target_name[len(ldmodule_prefix):]

    if target_name.endswith(ldmodule_suffix):
        target_name = target_name[:-len(ldmodule_suffix)]

    return target_name


def _ldmodule_soversion(target, source, env, for_signature):
    """Function to determine what to use for SOVERSION"""

    if 'SOVERSION' in env:
        return '.$SOVERSION'
    elif 'LDMODULEVERSION' in env:
        ldmod_version = env.subst('$LDMODULEVERSION')
        # We use only the most significant digit of LDMODULEVERSION
        return '.' + ldmod_version.split('.')[0]
    else:
        return ''


def _ldmodule_soname(target, source, env, for_signature) -> str:
    if 'SONAME' in env:
        return '$SONAME'
    else:
        return "$LDMODULEPREFIX$_get_ldmodule_stem${LDMODULESUFFIX}$_LDMODULESOVERSION"

def _LDMODULEVERSION(target, source, env, for_signature):
    """
    Return "." + version if it's set, otherwise just a blank
    """
    value = env.subst('$LDMODULEVERSION', target=target, source=source)
    # print("_has_LDMODULEVERSION:%s"%value)
    if value:
        return "."+value
    else:
        return ""

def setup_loadable_module_logic(env) -> None:
    """
    Just the logic for loadable modules

    For most platforms, a loadable module is the same as a shared
    library.  Platforms which are different can override these, but
    setting them the same means that LoadableModule works everywhere.

    :param env:
    :return:
    """

    createLoadableModuleBuilder(env)

    env['_get_ldmodule_stem'] = _get_ldmodule_stem
    env['_LDMODULESOVERSION'] = _ldmodule_soversion
    env['_LDMODULESONAME'] = _ldmodule_soname

    env['LDMODULENAME'] = '${LDMODULEPREFIX}$_get_ldmodule_stem${_LDMODULESUFFIX}'

    # This is the non versioned LDMODULE filename
    # If LDMODULEVERSION is defined then this will symlink to $LDMODULENAME
    env['LDMODULE_NOVERSION_SYMLINK'] = '$_get_shlib_dir${LDMODULEPREFIX}$_get_ldmodule_stem${LDMODULESUFFIX}'

    # This is the sonamed file name
    # If LDMODULEVERSION is defined then this will symlink to $LDMODULENAME
    env['LDMODULE_SONAME_SYMLINK'] = '$_get_shlib_dir$_LDMODULESONAME'

    env['_LDMODULEVERSION'] =  _LDMODULEVERSION

    env['_LDMODULEVERSIONFLAGS'] = '$LDMODULEVERSIONFLAGS -Wl,-soname=$_LDMODULESONAME'

    env['LDMODULEEMITTER'] = [lib_emitter, ldmod_symlink_emitter]

    env['LDMODULEPREFIX'] = '$SHLIBPREFIX'
    env['_LDMODULESUFFIX'] = '${LDMODULESUFFIX}${_LDMODULEVERSION}'
    env['LDMODULESUFFIX'] = '$SHLIBSUFFIX'

    env['LDMODULE'] = '$SHLINK'

    env['LDMODULEFLAGS'] = '$SHLINKFLAGS'

    env['LDMODULECOM'] = '$LDMODULE -o $TARGET $LDMODULEFLAGS $__LDMODULEVERSIONFLAGS $__RPATH $SOURCES ' \
                         '$_LIBDIRFLAGS $_LIBFLAGS '

    env['LDMODULEVERSION'] = '$SHLIBVERSION'
    env['LDMODULENOVERSIONSYMLINKS'] = '$SHLIBNOVERSIONSYMLINKS'


# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""CacheDir support
"""

import atexit
import json
import os
import shutil
import stat
import sys
import tempfile
import uuid

import SCons.Action
import SCons.Errors
import SCons.Warnings
import SCons.Util

CACHE_PREFIX_LEN = 2  # first two characters used as subdirectory name
CACHE_TAG = (
    b"Signature: 8a477f597d28d172789f06886806bc55\n"
    b"# SCons cache directory - see https://bford.info/cachedir/\n"
)

cache_enabled = True
cache_debug = False
cache_force = False
cache_show = False
cache_readonly = False
cache_tmp_uuid = uuid.uuid4().hex

def CacheRetrieveFunc(target, source, env) -> int:
    t = target[0]
    fs = t.fs
    cd = env.get_CacheDir()
    cd.requests += 1
    cachedir, cachefile = cd.cachepath(t)
    if not fs.exists(cachefile):
        cd.CacheDebug('CacheRetrieve(%s):  %s not in cache\n', t, cachefile)
        return 1
    cd.hits += 1
    cd.CacheDebug('CacheRetrieve(%s):  retrieving from %s\n', t, cachefile)
    if SCons.Action.execute_actions:
        if fs.islink(cachefile):
            fs.symlink(fs.readlink(cachefile), t.get_internal_path())
        else:
            cd.copy_from_cache(env, cachefile, t.get_internal_path())
            try:
                os.utime(cachefile, None)
            except OSError:
                pass
        st = fs.stat(cachefile)
        fs.chmod(t.get_internal_path(), stat.S_IMODE(st.st_mode) | stat.S_IWRITE)
    return 0

def CacheRetrieveString(target, source, env) -> str:
    t = target[0]
    cd = env.get_CacheDir()
    cachedir, cachefile = cd.cachepath(t)
    if t.fs.exists(cachefile):
        return "Retrieved `%s' from cache" % t.get_internal_path()
    return ""

CacheRetrieve = SCons.Action.Action(CacheRetrieveFunc, CacheRetrieveString)

CacheRetrieveSilent = SCons.Action.Action(CacheRetrieveFunc, None)

def CachePushFunc(target, source, env) -> None:
    if cache_readonly:
        return

    t = target[0]
    if t.nocache:
        return
    fs = t.fs
    cd = env.get_CacheDir()
    cachedir, cachefile = cd.cachepath(t)
    if fs.exists(cachefile):
        # Don't bother copying it if it's already there.  Note that
        # usually this "shouldn't happen" because if the file already
        # existed in cache, we'd have retrieved the file from there,
        # not built it.  This can happen, though, in a race, if some
        # other person running the same build pushes their copy to
        # the cache after we decide we need to build it but before our
        # build completes.
        cd.CacheDebug('CachePush(%s):  %s already exists in cache\n', t, cachefile)
        return

    cd.CacheDebug('CachePush(%s):  pushing to %s\n', t, cachefile)

    tempfile = "%s.tmp%s"%(cachefile,cache_tmp_uuid)
    errfmt = "Unable to copy %s to cache. Cache file is %s"

    try:
        fs.makedirs(cachedir, exist_ok=True)
    except OSError:
        msg = errfmt % (str(target), cachefile)
        raise SCons.Errors.SConsEnvironmentError(msg)
    try:
        if fs.islink(t.get_internal_path()):
            fs.symlink(fs.readlink(t.get_internal_path()), tempfile)
        else:
            cd.copy_to_cache(env, t.get_internal_path(), tempfile)
        fs.rename(tempfile, cachefile)

    except OSError:
        # It's possible someone else tried writing the file at the
        # same time we did, or else that there was some problem like
        # the CacheDir being on a separate file system that's full.
        # In any case, inability to push a file to cache doesn't affect
        # the correctness of the build, so just print a warning.
        msg = errfmt % (str(t), cachefile)
        cd.CacheDebug(errfmt + '\n', str(t), cachefile)
        SCons.Warnings.warn(SCons.Warnings.CacheWriteErrorWarning, msg)

CachePush = SCons.Action.Action(CachePushFunc, None)


class CacheDir:

    def __init__(self, path) -> None:
        """Initialize a CacheDir object.

        The cache configuration is stored in the object. It
        is read from the config file in the supplied path if
        one exists,  if not the config file is created and
        the default config is written, as well as saved in the object.
        """
        self.requests = 0
        self.hits = 0
        self.path = path
        self.current_cache_debug = None
        self.debugFP = None
        self.config = {}
        if path is not None:
            self._readconfig(path)

    def _add_config(self, path: str) -> None:
        """Create the cache config file in *path*.

        Locking isn't necessary in the normal case - when the cachedir is
        being created - because it's written to a unique directory first,
        before the directory is renamed. But it is legal to call CacheDir
        with an existing directory, which may be missing the config file,
        and in that case we do need locking. Simpler to always lock.
        """
        config_file = os.path.join(path, 'config')
        # TODO: this breaks the "unserializable config object" test which
        #   does some crazy stuff, so for now don't use setdefault. It does
        #   seem like it would be better to preserve an exisiting value.
        # self.config.setdefault('prefix_len', CACHE_PREFIX_LEN)
        self.config['prefix_len'] = CACHE_PREFIX_LEN
        with SCons.Util.FileLock(config_file, timeout=5, writer=True), open(
            config_file, "x"
        ) as config:
            try:
                json.dump(self.config, config)
            except Exception:
                msg = "Failed to write cache configuration for " + path
                raise SCons.Errors.SConsEnvironmentError(msg)

        # Add the tag file "carelessly" - the contents are not used by SCons
        # so we don't care about the chance of concurrent writes.
        try:
            tagfile = os.path.join(path, "CACHEDIR.TAG")
            with open(tagfile, 'xb') as cachedir_tag:
                cachedir_tag.write(CACHE_TAG)
        except FileExistsError:
            pass

    def _mkdir_atomic(self, path: str) -> bool:
        """Create cache directory at *path*.

        Uses directory renaming to avoid races.  If we are actually
        creating the dir, populate it with the metadata files at the
        same time as that's the safest way. But it's not illegal to point
        CacheDir at an existing directory that wasn't a cache previously,
        so we may have to do that elsewhere, too.

        Returns:
            ``True`` if it we created the dir, ``False`` if already existed,

        Raises:
            SConsEnvironmentError: if we tried and failed to create the cache.
        """
        directory = os.path.abspath(path)
        if os.path.exists(directory):
            return False

        try:
            # TODO: Python 3.7. See comment below.
            # tempdir = tempfile.TemporaryDirectory(dir=os.path.dirname(directory))
            tempdir = tempfile.mkdtemp(dir=os.path.dirname(directory))
        except OSError as e:
            msg = "Failed to create cache directory " + path
            raise SCons.Errors.SConsEnvironmentError(msg) from e

        # TODO: Python 3.7: the context manager raises exception on cleanup
        #    if the temporary was moved successfully (File Not Found).
        #    Fixed in 3.8+. In the replacement below we manually clean up if
        #    the move failed as mkdtemp() does not. TemporaryDirectory's
        #    cleanup is more sophisitcated so prefer when we can use it.
        # self._add_config(tempdir.name)
        # with tempdir:
        #     try:
        #         os.replace(tempdir.name, directory)
        #         return True
        #     except OSError as e:
        #         # did someone else get there first?
        #         if os.path.isdir(directory):
        #             return False  # context manager cleans up
        #         msg = "Failed to create cache directory " + path
        #         raise SCons.Errors.SConsEnvironmentError(msg) from e

        self._add_config(tempdir)
        try:
            os.replace(tempdir, directory)
            return True
        except OSError as e:
            # did someone else get there first? attempt cleanup.
            if os.path.isdir(directory):
                try:
                    shutil.rmtree(tempdir)
                except Exception:  # we tried, don't worry about it
                    pass
                return False
            msg = "Failed to create cache directory " + path
            raise SCons.Errors.SConsEnvironmentError(msg) from e

    def _readconfig(self, path: str) -> None:
        """Read the cache config from *path*.

        If directory or config file do not exist, create and populate.
        """
        config_file = os.path.join(path, 'config')
        created = self._mkdir_atomic(path)
        if not created and not os.path.isfile(config_file):
            # Could have been passed an empty directory
            self._add_config(path)
        try:
            with SCons.Util.FileLock(config_file, timeout=5, writer=False), open(
                config_file
            ) as config:
                self.config = json.load(config)
        except (ValueError, json.decoder.JSONDecodeError):
            msg = "Failed to read cache configuration for " + path
            raise SCons.Errors.SConsEnvironmentError(msg)

    def CacheDebug(self, fmt, target, cachefile) -> None:
        if cache_debug != self.current_cache_debug:
            if cache_debug == '-':
                self.debugFP = sys.stdout
            elif cache_debug:
                def debug_cleanup(debugFP) -> None:
                    debugFP.close()

                self.debugFP = open(cache_debug, 'w')
                atexit.register(debug_cleanup, self.debugFP)
            else:
                self.debugFP = None
            self.current_cache_debug = cache_debug
        if self.debugFP:
            self.debugFP.write(fmt % (target, os.path.split(cachefile)[1]))
            self.debugFP.write("requests: %d, hits: %d, misses: %d, hit rate: %.2f%%\n" %
                               (self.requests, self.hits, self.misses, self.hit_ratio))

    @classmethod
    def copy_from_cache(cls, env, src, dst) -> str:
        """Copy a file from cache."""
        if env.cache_timestamp_newer:
            return env.fs.copy(src, dst)
        else:
            return env.fs.copy2(src, dst)

    @classmethod
    def copy_to_cache(cls, env, src, dst) -> str:
        """Copy a file to cache.

        Just use the FS copy2 ("with metadata") method, except do an additional
        check and if necessary a chmod to ensure the cachefile is writeable,
        to forestall permission problems if the cache entry is later updated.
        """
        try:
            result = env.fs.copy2(src, dst)
            st = stat.S_IMODE(os.stat(result).st_mode)
            if not st | stat.S_IWRITE:
                os.chmod(dst, st | stat.S_IWRITE)
            return result
        except AttributeError as ex:
            raise OSError from ex

    @property
    def hit_ratio(self) -> float:
        return (100.0 * self.hits / self.requests if self.requests > 0 else 100)

    @property
    def misses(self) -> int:
        return self.requests - self.hits

    def is_enabled(self) -> bool:
        return cache_enabled and self.path is not None

    def is_readonly(self) -> bool:
        return cache_readonly

    def get_cachedir_csig(self, node) -> str:
        cachedir, cachefile = self.cachepath(node)
        if cachefile and os.path.exists(cachefile):
            return SCons.Util.hash_file_signature(cachefile, SCons.Node.FS.File.hash_chunksize)

    def cachepath(self, node) -> tuple:
        """Return where to cache a file.

        Given a Node, obtain the configured cache directory and
        the path to the cached file, which is generated from the
        node's build signature. If caching is not enabled for the
        None, return a tuple of None.
        """
        if not self.is_enabled():
            return None, None

        sig = node.get_cachedir_bsig()
        subdir = sig[:self.config['prefix_len']].upper()
        cachedir = os.path.join(self.path, subdir)
        return cachedir, os.path.join(cachedir, sig)

    def retrieve(self, node) -> bool:
        """Retrieve a node from cache.

        Returns True if a successful retrieval resulted.

        This method is called from multiple threads in a parallel build,
        so only do thread safe stuff here. Do thread unsafe stuff in
        built().

        Note that there's a special trick here with the execute flag
        (one that's not normally done for other actions).  Basically
        if the user requested a no_exec (-n) build, then
        SCons.Action.execute_actions is set to 0 and when any action
        is called, it does its showing but then just returns zero
        instead of actually calling the action execution operation.
        The problem for caching is that if the file does NOT exist in
        cache then the CacheRetrieveString won't return anything to
        show for the task, but the Action.__call__ won't call
        CacheRetrieveFunc; instead it just returns zero, which makes
        the code below think that the file *was* successfully
        retrieved from the cache, therefore it doesn't do any
        subsequent building.  However, the CacheRetrieveString didn't
        print anything because it didn't actually exist in the cache,
        and no more build actions will be performed, so the user just
        sees nothing.  The fix is to tell Action.__call__ to always
        execute the CacheRetrieveFunc and then have the latter
        explicitly check SCons.Action.execute_actions itself.
        """
        if not self.is_enabled():
            return False

        env = node.get_build_env()
        if cache_show:
            if CacheRetrieveSilent(node, [], env, execute=1) == 0:
                node.build(presub=0, execute=0)
                return True
        else:
            if CacheRetrieve(node, [], env, execute=1) == 0:
                return True

        return False

    def push(self, node):
        if self.is_readonly() or not self.is_enabled():
            return
        return CachePush(node, [], node.get_build_env())

    def push_if_forced(self, node):
        if cache_force:
            return self.push(node)

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:
